/**
 * Clam 相关埋点
 */
const http = require('http');
const url = require('url');
const os = require('os');
const path = require('path');
const querystring = require('querystring');
const shell = require('shelljs');
const PWD = process.cwd();
let pkgInfo = {};

const GOLDLOG_HOST = 'gm.mmstat.com';

// For charles debug
// const GOLDLOG_HOST = 'localhost:54818';

/**
 * 黄金令箭打点
 * @param {String} prefix 
 * @param {Object} argObj 
 * @param {Function} callback 
 */
const glodlog = function (prefix, argObj, callback) {
  // 获取项目信息
  try {
    pkgInfo = require(path.join(PWD, 'package.json'));
  } catch (e) {}
  // 带上环境信息
  let userName = ''; 
  let userEmail = '';
  let userId = process.env.BUILD_USER || '';
  try {
    const userNameGitConfig = shell.exec('git config --global user.name', {
      silent: true
    });
    if (userNameGitConfig.code == 0) {
      userName = userNameGitConfig.output.toString().trim();
    }
    const userEmailGitConfig = shell.exec('git config --global user.email', {
      silent: true
    });
    if (userEmailGitConfig.code == 0) {
      userEmail = userEmailGitConfig.output.toString().trim();
    }
  } catch (e) {}
  const cpus = os.cpus();
  const mergedArgObj = Object.assign({
    nodeVersion: process.versions.node,
    platform: process.platform,
    user: userEmail,
    userid: userId,
    username: encodeURIComponent(userName),
    osVersion: os.release(),
    cpuCoreNum: cpus.length,
    cpuModel: cpus[0].model,
    pkgName: pkgInfo.name,
    pkgVersion: pkgInfo.version,
    pkgGroup: pkgInfo.group,
    pkgLib: pkgInfo.lib,
    pkgType: pkgInfo.type,
    pkgMobile: pkgInfo.mobile,
    clamVersion: pkgInfo.clamVersion,
    commandType: pkgInfo.commandType || 'grunt',
    buildEnv: process.env.BUILD_ENV === 'cloud' ? 'cloud' : 'local',
    builderName: process.env.CLAM_BUILDER_NAME || '',
    builderVersion: process.env.CLAM_BUILDER_VERSION || ''
  }, argObj);

  const reqUrl = `http://${GOLDLOG_HOST}/${prefix}?t=${Date.now()}&${querystring.stringify(mergedArgObj)}`;
  http.get(Object.assign(url.parse(reqUrl), {
    timeout: 1000 * 5,  // 5s timeout
    agent: false        // create a new agent just for this one request
  }), callback || (() => {}));
};

exports.glodlog = glodlog;

/**
 * `time-grunt` 回调
 */
exports.logGruntTime = function (stats, done) {
  const TASK_MAP = {};
  let totalTime = 0;
  stats.forEach((stat) => {
   TASK_MAP[encodeURIComponent(stat[0])] = stat[1];
   totalTime += +stat[1];
  });
  glodlog('clam.grunt-build-perf.grunt-build-time', {
    stats: JSON.stringify(TASK_MAP),
    tasks: encodeURIComponent(this.cli.tasks.join()),
    totalTime,
  }, done);
};
